
#include <iostream>
#include <string>

#include "ClientDataManager.h"
#include "ClientNetworkNode.h"
#include "packetID.h"
#include "OgreConsole.h"
#include "IntroState.h"
#include "PlayState.h"
#include "InputManager.h"
#include "Enemy.h"
#include "GameOverState.h"

using namespace Ogre;

// ============================================================================
void ClientDataManager::processInput(unsigned long timeSinceInMilis)
{
	if(!mLocalPlayer.isInitialized())
		return;


	// This whole processing is also a part of the Dead Reckoning system, because
	// after processInput we call ClientDataManager::update() which writes the real
	// player data (including local player) to the classes.
	// So, if the server is working 100% perfect (no packets lost, 0 ms ping etc)
	// we wouldn't need to update the position of the local player.
	// However we assume some packets can be lost, so we update our position as well.

	OIS::Keyboard * keyboard = InputManager::getSingleton().getKeyboard();

	// Accelerate/deaccelerate
	if(keyboard->isKeyDown(OIS::KC_W) || keyboard->isKeyDown(OIS::KC_S))
	{
		float accelDeaccel = keyboard->isKeyDown(OIS::KC_W)? -1.0 : 1.0f;

		mLocalPlayer.processAccceleration(timeSinceInMilis, accelDeaccel);
		mLocalPlayer.processVelocity(timeSinceInMilis);
		mLocalPlayer.processPosition(timeSinceInMilis);
	}
	else
	{
		// brakes!
		float relativeSpeed = mLocalPlayer.getVelocity().length() / mLocalPlayer.mMaxSpeed;
		if(relativeSpeed > 0.05f) // greater than 5% of max speed
		{
			Vector3 currVelocityDirection = mLocalPlayer.getVelocity().normalisedCopy();
			
			mLocalPlayer.setAcceleration( -mLocalPlayer.mMaxAcceleration * relativeSpeed * 0.8f * currVelocityDirection);
		}
		else
		{ // don't bother to use brakes
			mLocalPlayer.setAcceleration( Vector3(0,0,0) );
		}
		mLocalPlayer.processVelocity(timeSinceInMilis);
		mLocalPlayer.processPosition(timeSinceInMilis);
	}

	int mouseRelativeX = InputManager::getSingleton().getMouse()->getMouseState().X.rel;
	int mouseRelativeY = InputManager::getSingleton().getMouse()->getMouseState().Y.rel;

	// Rotation brake
	if((!keyboard->isKeyDown(OIS::KC_A) && !keyboard->isKeyDown(OIS::KC_D)
		&& !keyboard->isKeyDown(OIS::KC_Q) && !keyboard->isKeyDown(OIS::KC_E)
		&& mouseRelativeX == 0 && mouseRelativeY == 0)
		|| keyboard->isKeyDown(OIS::KC_R)) // rotation brake!
	{
		mLocalPlayer.processRotationBrake(timeSinceInMilis);
	}

	// Yaw
	float yawMultiplier = 0.0f;
	if(keyboard->isKeyDown(OIS::KC_D) && keyboard->isKeyDown(OIS::KC_A) == false)
	{
		yawMultiplier = -1.0f;
	}
	if(keyboard->isKeyDown(OIS::KC_A) && keyboard->isKeyDown(OIS::KC_D) == false)
	{
		yawMultiplier = 1.0f;
	}
	if(mouseRelativeX != 0)
	{
		//OgreConsole::getSingleton().print(Ogre::StringConverter::toString( mouseRelativeX ) );
		if(mouseRelativeX > 0)
		{
			yawMultiplier = (float)mouseRelativeX * -0.03;
			//OgreConsole::getSingleton().print(Ogre::StringConverter::toString( yawMultiplier ) );
		}
		else
		{
			yawMultiplier = (float)mouseRelativeX * -0.03;
			//OgreConsole::getSingleton().print(Ogre::StringConverter::toString( yawMultiplier ) );
		}
	}
	mLocalPlayer.processYawChange(timeSinceInMilis, yawMultiplier);
	mLocalPlayer.processYaw(timeSinceInMilis);

	// Pitch
	float pitchMultiplier = 0.0f;
	if(keyboard->isKeyDown(OIS::KC_E) && keyboard->isKeyDown(OIS::KC_Q) == false)
	{
		pitchMultiplier = 1.0f;
	}
	if(keyboard->isKeyDown(OIS::KC_Q) && keyboard->isKeyDown(OIS::KC_E) == false)
	{
		pitchMultiplier = -1.0f;
	}
	if(mouseRelativeY != 0)
	{
		//OgreConsole::getSingleton().print(Ogre::StringConverter::toString( mouseRelativeY ) );
		if(mouseRelativeY > 0)
		{
			pitchMultiplier = (float)mouseRelativeY * -0.03;
			//OgreConsole::getSingleton().print(Ogre::StringConverter::toString( pitchMultiplier ) );
		}
		else
		{
			pitchMultiplier = (float)mouseRelativeY * -0.03;
			//OgreConsole::getSingleton().print(Ogre::StringConverter::toString( pitchMultiplier ) );
		}
	}
	mLocalPlayer.processPitchChange(timeSinceInMilis, pitchMultiplier);
	mLocalPlayer.processPitch(timeSinceInMilis);

	// Fire missiles!
	mLocalPlayer.mMilisSinceLastFiring += timeSinceInMilis;
	if( ( keyboard->isKeyDown(OIS::KC_SPACE) || 
		  InputManager::getSingleton().getMouse()->getMouseState().buttonDown(OIS::MB_Left)
		 )
		&& mLocalPlayer.mMilisSinceLastFiring >= mLocalPlayer.mFiringDelayInMilis
	  )
	{
		mLocalPlayer.mMilisSinceLastFiring = 0;
		isPlayerFiringValid = true;
	}

}
// ============================================================================
void ClientDataManager::update(unsigned long timeSinceInMilis)
{
	// Update the player
	if(mLocalPlayer.isInitialized())
	{
		// WARNING! This is not strict server/client
		// plus someone can easily cheat here 

		// Strict server/client = send only keyboard state/message requests (ie what a player WANTS to do)

		// Check whether we're still alive
		if(mLocalPlayer.getLives() < 1)
		{
			// Oops
			this->gameOver();
			return;
		}

		// Send our player data:
		this->timeSinceLastPlayerPacketUpdateInMilis += timeSinceInMilis;
		if(timeSinceLastPlayerPacketUpdateInMilis >= 66) // roughly 1/15th of a second
		{
			char playerDataBuffer[ID_PlayerData_Size];
			mLocalPlayer.saveDataToPacket(playerDataBuffer);
			mNetworkNode.sendToServer(playerDataBuffer,ID_PlayerData_Size);

			timeSinceLastPlayerPacketUpdateInMilis = 0;
		}

		//sending a fire missile request, and thats it - the server will handle the rest
		if(isPlayerFiringValid == true)
		{
			// This variable is set during the processInput stage
			isPlayerFiringValid = false;

			char data[ID_PlayerRequestsCreateMissile_Size];
			memset(data,0,ID_PlayerRequestsCreateMissile_Size);
			data[0] = ID_PlayerRequestsCreateMissile;
			memcpy(&data[1], mLocalPlayer.getPlayerName().c_str(), mLocalPlayer.getPlayerName().length());

			mNetworkNode.sendToServer(data,ID_PlayerRequestsCreateMissile_Size);
		}
	}

	// Correct the errors and collect fresh data:
	sockaddr_in serverAddress;

	while(mNetworkNode.hasMorePackets())
	{
		// clear previous data
		memset(&serverAddress,0,sizeof(sockaddr_in));
		memset(lastReadDataBuf,0,MAXBUFFERSIZE);

		lastNumberOfBytesRead = mNetworkNode.readPacket( &serverAddress, lastReadDataBuf, MAXBUFFERSIZE);
		
		// ================= main packet handling ====================
		if(lastNumberOfBytesRead < 0)
		{
			std::cerr << "ClientNetworkNode error: ";
		}
		else
		{
			switch(lastReadDataBuf[0])
			{
				case ID_ConnectionAccepted:
					// NetworkNode needs to accept the connection etc
					mNetworkNode.handlePacketConnectionAccepted(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					// DataManager will store the data (like accepted PlayerName)
					this->handlePacketConnectionAccepted(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_SendNewPlayerData:
					this->handlePacketSendNewPlayerData(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_ServerPlayerData:
					this->handlePacketServerPlayerData(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_ServerCreatesEnemy:
					this->handlePacketCreateEnemy(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_EnemyData:
					this->handlePacketEnemyData(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_ServerNotifiesAboutPlayerExit:
					this->handlePacketServerNotifiesAboutPlayerExit(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_ServerCreatesMissile:
					this->handlePacketServerCreatesMissile(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_MissileData:
					this->handlePacketMissileData(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_ServerDestroysEnemy:
					this->handlePacketServerDestroysEnemy(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_ServerChangesPlayerScore:
					this->handlePacketServerChangesPlayerScore(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_ServerDestroysMissile:
					this->handlePacketID_ServerDestroysMissile(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
					break;
				case ID_ServerChangesPlayerLives:
					this->handlePacketServerChangesPlayerLives(&serverAddress,lastReadDataBuf,
						lastNumberOfBytesRead);
				default:
					OgreConsole::getSingleton().print("Received unhandled data: " 
						+std::string(lastReadDataBuf) );
					break;
			}
		}
		// ===========================================================
	}

	// DEAD RECKONING
	if(this->isDeadReckoningActive)
	{
		// Interpolate other clients - DEAD RECKONING
		// for each player..
		map<string, Player>::iterator iter;
		for(iter = mOtherPlayers.begin(); iter != mOtherPlayers.end(); iter++)
		{
			Player * player = &(*iter).second;

			player->processVelocity(timeSinceInMilis);
			player->processPosition(timeSinceInMilis);
			player->processPitch(timeSinceInMilis);
			player->processRoll(timeSinceInMilis);
			player->processYaw(timeSinceInMilis);
		}

		// Interpolate enemies - DEAD RECKONING
		EnemyMapType::iterator enemyProcessingIter;
		for(enemyProcessingIter = mEnemies.begin(); enemyProcessingIter != mEnemies.end();
				++enemyProcessingIter)
		{
			Enemy * enemy = &(*enemyProcessingIter).second;
			enemy->processAccceleration(timeSinceInMilis, -1.0f);
			enemy->processVelocity(timeSinceInMilis);
			enemy->processPosition(timeSinceInMilis);
		}

		// Interpolate missiles - DEAD RECKONING
		MissileMapType::iterator missileProcessingIter;
		for(missileProcessingIter = mMissiles.begin(); missileProcessingIter != mMissiles.end();
				++missileProcessingIter)
		{
			Missile * missile = &(*missileProcessingIter).second;
			missile->processAccceleration(timeSinceInMilis, -1.0f);
			missile->processVelocity(timeSinceInMilis);
			missile->processPosition(timeSinceInMilis);
		}
	}
}
// ============================================================================
void ClientDataManager::handlePacketMissileData(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	string missileId(&data[1]);

	MissileMapType::iterator iter;
	iter = mMissiles.find(missileId);
	if(iter != mMissiles.end())
	{
		(*iter).second.readDataFromPacket(data);
	}
	else
	{
		// Enemy known? Packet loss?
		char response[ID_UnknownEntity_Size];
		memset(response,0,ID_UnknownEntity_Size);
		response[0] = ID_UnknownEntity;
		memcpy(&response[1],missileId.c_str(),MAXENTITYIDSIZE);

		mNetworkNode.sendToServer(response,ID_UnknownEntity_Size);
	}
}
// ============================================================================
void ClientDataManager::handlePacketServerCreatesMissile(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	Missile missile;
	missile.initializeFromPacket(data,"PlayState SceneManager");
	missile.mMesh->setVisible(true);

	mMissiles.insert(make_pair(missile.getId(),missile));
}
// ============================================================================
void ClientDataManager::handlePacketServerNotifiesAboutPlayerExit(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	std::string exitingPlayerName(&data[1]/*skip packetID */);

	std::map<std::string, Player>::iterator playerIter;
	playerIter = mOtherPlayers.find(exitingPlayerName);
	if(playerIter != mOtherPlayers.end())
	{
		// mMesh, mSceneNode and mParticleNode/Sys
		playerIter->second.mSceneNode->detachAllObjects();

		Root::getSingleton().getSceneManager(
		"PlayState SceneManager")->destroyEntity(playerIter->second.mMesh->getName());

		Root::getSingleton().getSceneManager(
			"PlayState SceneManager")->destroyParticleSystem(playerIter->second.mParticleSys);

		playerIter->second.mSceneNode->removeChild( playerIter->second.mParticleNode );

		Ogre::SceneNode* todestroy = (Ogre::SceneNode*)playerIter->second.mSceneNode->getParent()->removeChild( playerIter->second.mSceneNode->getName() );
		
		Root::getSingleton().getSceneManager(
			"PlayState SceneManager")->destroySceneNode(todestroy);

		// Remove physical class instance
		mOtherPlayers.erase(playerIter);
	}
}
// ============================================================================
void ClientDataManager::handlePacketEnemyData(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	string enemyId(&data[1]);

	EnemyMapType::iterator iter;
	iter = mEnemies.find(enemyId);
	if(iter != mEnemies.end())
	{
		(*iter).second.readDataFromPacket(data);
	}
	else
	{
		// We don't know this enemy yet. Packet loss?
		char response[ID_UnknownEntity_Size];
		memset(response,0,ID_UnknownEntity_Size);
		response[0] = ID_UnknownEntity;
		memcpy(&response[1],enemyId.c_str(),MAXENTITYIDSIZE);

		mNetworkNode.sendToServer(response,ID_UnknownEntity_Size);
	}

}
// ============================================================================
void ClientDataManager::handlePacketCreateEnemy(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	Enemy enemy;
	enemy.initializeFromPacket(data, "PlayState SceneManager");

	mEnemies.insert(make_pair(enemy.getId(), enemy));
}
// ============================================================================
void ClientDataManager::handlePacketSendNewPlayerData(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	if(!mNetworkNode.isConnected())
		return;

	if(PlayState::getSingletonPtr()->isActive() == false)
	{
		// START!!
		IntroState::getSingletonPtr()->changeState(PlayState::getSingletonPtr());
	}

	int cnt = 1;
	cnt += MAXENTITYIDSIZE;
	string playerName(&data[cnt]);

	if(playerName.compare(mPlayerName) == 0)
	{
		// Thats us!
		Player player;
		player.initializeFromPacket(data,"PlayState SceneManager",true);
		mLocalPlayer = player;
	}
	else
	{
		// Someone else
		Player player;
		player.initializeFromPacket(data,"PlayState SceneManager");
		mOtherPlayers.insert(make_pair(player.getPlayerName(),player));
	}
}
// ============================================================================
void ClientDataManager::handlePacketServerPlayerData(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	if(mLocalPlayer.isInitialized() == false)
	{
		return;
	}

	string playerName(&data[1]);

	if(playerName.compare(this->mPlayerName) == 0)
	{
		// hes talking about us
		mLocalPlayer.readDataFromPacket(data);
	}
	else
	{
		// someone else
		map<string, Player>::iterator iter;
		iter = mOtherPlayers.find(playerName);
		if(iter != mOtherPlayers.end())
		{
			(*iter).second.readDataFromPacket(data);
			return;
		}
	}
	
}
// ============================================================================
void ClientDataManager::handlePacketConnectionAccepted(
								sockaddr_in * senderAddress,
								const char data[],
								const int dataSize)
{
	if(!mNetworkNode.isConnected())
		return;

	// we can skip packetID:
	//char packetID = data[0];

	// we can skip the colour data for now
	// data[1]

	// get the accepted player name
	mPlayerName = std::string(data+2,dataSize-2);

	// debug:
	OgreConsole::getSingleton().print("Client: server ("
		+std::string( inet_ntoa(senderAddress->sin_addr) )
		+") accepted connection as '"
		+mPlayerName
		+"'"
		);
}
// ============================================================================
void ClientDataManager::connect(string address, string playername)
{
	mNetworkNode.connect(address, playername);
}
// ============================================================================
void ClientDataManager::init()
{
	if(!initialized)
	{
		mNetworkNode.start();
		initialized = true;
	}
}
// ============================================================================
ClientDataManager & ClientDataManager::getSingleton()
{
	static ClientDataManager instance;
	return instance;
}
// ============================================================================
void ClientDataManager::disconnect()
{
	if(mNetworkNode.isConnected())
	{
		mNetworkNode.disconnect( this->mPlayerName );
	}
}
// ============================================================================
void ClientDataManager::handlePacketServerDestroysEnemy(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	String entityId(&data[1]);

	EnemyMapType::iterator enemyProcessingIter = mEnemies.find(entityId);
	if(enemyProcessingIter != mEnemies.end())
	{
		Enemy * enemy = &(*enemyProcessingIter).second;

		// Destroy locally
		enemy->destroy(String("PlayState SceneManager"));
		mEnemies.erase(enemyProcessingIter);
	}	
}
// ============================================================================
void ClientDataManager::handlePacketServerChangesPlayerScore(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	string playerName(&data[1]);

	if(playerName.compare(this->mPlayerName) == 0)
	{
		// hes talking about us
		mLocalPlayer.readScoreFromPacket(data);
	}
	else
	{
		// someone else
		map<string, Player>::iterator iter;
		iter = mOtherPlayers.find(playerName);
		if(iter != mOtherPlayers.end())
		{
			(*iter).second.readScoreFromPacket(data);
			return;
		}
	}
}
// ============================================================================
void ClientDataManager::handlePacketID_ServerDestroysMissile(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	String entityId(&data[1]);

	MissileMapType::iterator missileProcessingIter = mMissiles.find(entityId);
	if(missileProcessingIter != mMissiles.end())
	{
		Missile * missile = &(*missileProcessingIter).second;

		// Destroy locally
		missile->destroy(String("PlayState SceneManager"));
		mMissiles.erase(missileProcessingIter);
	}	
}
// ============================================================================
void ClientDataManager::handlePacketServerChangesPlayerLives(
									sockaddr_in * senderAddress, 
									const char data[],
								   const int dataSize)
{
	string playerName(&data[1]);

	if(playerName.compare(this->mPlayerName) == 0)
	{
		// hes talking about us
		mLocalPlayer.readLivesFromPacket(data);
	}
	else
	{
		// someone else
		map<string, Player>::iterator iter;
		iter = mOtherPlayers.find(playerName);
		if(iter != mOtherPlayers.end())
		{
			(*iter).second.readLivesFromPacket(data);
			return;
		}
	}
}
// ============================================================================
void ClientDataManager::gameOver()
{
	mOtherPlayers.clear();
	mEnemies.clear();
	mMissiles.clear();

	this->disconnect();

	mLocalPlayer.mIsInitialized = false;

	PlayState::getSingletonPtr()->changeState(GameOverState::getSingletonPtr());
}
// ============================================================================
void ClientDataManager::reconnect()
{
	mNetworkNode.reconnect(mLocalPlayer.getPlayerName());
}
// ============================================================================