
#include <OgreException.h>

#include "GameManager.h"
#include "GameState.h"
#include "GuiManager.h"

using namespace Ogre;

// ============================================================================
GameManager::GameManager() :
    mRoot( 0 ),
    bShutdown( false )
{
	// nuthin'
}
// ============================================================================
GameManager::~GameManager()
{
	if( mRoot )
	{
        delete mRoot;
        mRoot = 0;
    }
}
// ============================================================================
void GameManager::startGame( GameState *gameState)
{
  	createRoot();
	defineResources();
	setupRenderSystem();
	createRenderWindow();

	initializeResourceGroups();
	setupInputSystem();

	this->changeState( gameState );

	 unsigned long lTimeLastFrame = 0;

	 while( !bShutdown )
	 {
        // Calculate time since last frame and remember current time for next frame
        unsigned long lTimeCurrentFrame = mRoot->getTimer()->getMilliseconds();
        unsigned long lTimeSinceLastFrame = lTimeCurrentFrame - lTimeLastFrame;
		lTimeSinceLastFrame = lTimeSinceLastFrame;
        lTimeLastFrame = lTimeCurrentFrame;

        // Update inputmanager
        InputManager::getSingleton().capture();

        // Update current state
        mStates.back()->update( lTimeSinceLastFrame );

		WindowEventUtilities::messagePump();//yys
		// Render next frame
        mRoot->renderOneFrame();
	 }

	 this->processShutdown();

	 // destroy!
}
// ============================================================================
void GameManager::changeState( GameState *gameState )
{
    // Cleanup the current state
    if( !mStates.empty() ) {
        mStates.back()->exit();
        mStates.pop_back();
    }

    // Store and init the new state
    mStates.push_back( gameState );
    mStates.back()->enter();
}
// ============================================================================
void GameManager::pushState( GameState *gameState )
{
    // Pause current state
    if( !mStates.empty() ) {
        mStates.back()->pause();
    }

    // Store and init the new state
    mStates.push_back( gameState );
    mStates.back()->enter();
}
// ============================================================================
void GameManager::popState()
{
    // Cleanup the current state
    if( !mStates.empty() ) {
        mStates.back()->exit();
        mStates.pop_back();
    }

    // Resume previous state
    if( !mStates.empty() ) {
        mStates.back()->resume();
    }
}
// ============================================================================
void GameManager::requestShutdown()
{
    bShutdown = true;
}
// ============================================================================
bool GameManager::keyPressed( const OIS::KeyEvent &e )
{
    // Call keyPressed of current state
    mStates.back()->keyPressed( e );

    return true;
}
// ============================================================================
bool GameManager::keyReleased( const OIS::KeyEvent &e )
{
    // Call keyReleased of current state
    mStates.back()->keyReleased( e );

    return true;
}
// ============================================================================
bool GameManager::mouseMoved( const OIS::MouseEvent &e )
{
    // Call mouseMoved of current state
    mStates.back()->mouseMoved( e );

    return true;
}
// ============================================================================
bool GameManager::mousePressed( const OIS::MouseEvent &e, OIS::MouseButtonID id )
{
    // Call mousePressed of current state
    mStates.back()->mousePressed( e, id );

    return true;
}
// ============================================================================
bool GameManager::mouseReleased( const OIS::MouseEvent &e, OIS::MouseButtonID id )
{
    // Call mouseReleased of current state
    mStates.back()->mouseReleased( e, id );

    return true;
}
// ============================================================================
GameManager & GameManager::getSingleton()
{
	static GameManager instance;
    return instance;
}
// ============================================================================
void GameManager::createRoot()
{
	mRoot = new Root();
}
// ============================================================================
void GameManager::defineResources()
{
	String secName, typeName, archName;
	ConfigFile cf;
	cf.load("resources.cfg");

	ConfigFile::SectionIterator seci = cf.getSectionIterator();
	
	while (seci.hasMoreElements())
	{
		secName = seci.peekNextKey();
		ConfigFile::SettingsMultiMap *settings = seci.getNext();
		ConfigFile::SettingsMultiMap::iterator i;
		for (i = settings->begin(); i != settings->end(); ++i)
		{
			typeName = i->first;
			archName = i->second;
			ResourceGroupManager::getSingleton().addResourceLocation(archName, typeName, secName);
		}
	}
}
// ============================================================================
void GameManager::setupRenderSystem()
{
	if (!mRoot->showConfigDialog())
		throw Exception(52, "User canceled the config dialog!", "GameManager::setupRenderSystem()");

}
// ============================================================================
void GameManager::createRenderWindow()
{
	 mRenderWindow = mRoot->initialise( true, "Ogre3DNetworkGame" );
}
// ============================================================================
void GameManager::initializeResourceGroups()
{
	//TextureManager::getSingleton().setDefaultNumMipmaps(5);
	ResourceGroupManager::getSingleton().initialiseAllResourceGroups();
}
// ============================================================================
void GameManager::setupScene()
{
	/* SceneManager * */  mgr = mRoot->createSceneManager(ST_EXTERIOR_CLOSE, "PlayState SceneManager");
	/* Camera * */  cam = mgr->createCamera("Camera");
	/* Viewport * */  vp = mRoot->getAutoCreatedWindow()->addViewport(cam);
	// mgr->setSkyBox(true,"Examples/SpaceSkyBox"); //todo

}
// ============================================================================
void GameManager::setupInputSystem()
{
	mRenderWindow = mRoot->getAutoCreatedWindow();
    InputManager::getSingleton().initialise( mRenderWindow );
	WindowEventUtilities::addWindowEventListener( mRenderWindow, this );//yys

    InputManager::getSingleton().addKeyListener( this, "GameManager" );
    InputManager::getSingleton().addMouseListener( this, "GameManager" );

}
// ============================================================================
void GameManager::setupLighting()
{
	SceneManager *mgr = mRoot->getSceneManager("PlayState SceneManager");
	mgr->setAmbientLight(ColourValue(1.0f, 1.0f, 1.0f));
}
// ============================================================================
void GameManager::startRenderLoop()
{
	mRoot->startRendering();

}
// ============================================================================
void GameManager::loadModel()
{

//	SceneManager *mgr = mRoot->getSceneManager("PlayState SceneManager");


//	Camera *cam = mgr->getCamera("Camera");


//	mPlayer = new Player();
	
	
	//playListener = new PlayerListener(mMouse, mKeyboard, mPlayer, cam);
//	mRoot->addFrameListener(playListener);

}
// ============================================================================
bool GameManager::windowClosing(RenderWindow* rw)
{
	requestShutdown();
	return true;
}
// ============================================================================
void GameManager::processShutdown()
{
	// Clean up all the states
    while( !mStates.empty() )
	{
        mStates.back()->exit();
        mStates.pop_back();
    }
}