
#include <Ogre.h>
#include <string.h>
#include <Winsock2.h>

#include "Missile.h"
#include "packetID.h"
#include "ParsingUtils.h"

using namespace Ogre;

// ============================================================================
Missile::Missile(void) :
	mMesh(0),
	mSceneNode(0),
	mMaxSpeed(15),
	mMaxAcceleration(5),
	mId(""),
	mClassName("Missile"),
	mIsInitialized(false),
	mCurrentFlyingTimeInMillis(0)
{
}
// ============================================================================
Missile::~Missile(void)
{
}
// ============================================================================
void Missile::destroy(Ogre::String & sceneManagerName)
{
	SceneManager * sceneMgr = Root::getSingleton().getSceneManager(sceneManagerName);
	
	// mMesh, mSceneNode
	mSceneNode->detachAllObjects();

	sceneMgr->destroyEntity(mMesh->getName());

	// This is optional to include missile trails in the future 
	//sceneMgr->destroyParticleSystem( mParticleSys );
	//mSceneNode->removeChild( mParticleNode );

	SceneNode* todestroy = (SceneNode*)mSceneNode->getParent()->removeChild( mSceneNode->getName());

	sceneMgr->destroySceneNode(todestroy);
}
// ============================================================================
void Missile::setOwner(Ogre::String _owner)
{
	mOwner = _owner;
}
// ============================================================================
void Missile::setId(Ogre::String _id)
{
	mId = _id;
}
// ============================================================================
void Missile::setVelocity(Ogre::Vector3 & _velocity)
{
	mVelocity = _velocity;
}
// ============================================================================
void Missile::setAcceleration(Ogre::Vector3 & _acceleration)
{
	mAcceleration = _acceleration;
}
// ============================================================================
Missile & Missile::operator=(Missile & _other)
{
	mMesh = _other.mMesh;
	mSceneNode = _other.mSceneNode;

	mOwner = _other.mOwner;
	mId = _other.mId;

	mVelocity = _other.mVelocity;
	mAcceleration = _other.mAcceleration;

	mIsInitialized = _other.mIsInitialized;

	return *this;
}
// ============================================================================
void Missile::setPosition(Ogre::Vector3 & _position)
{
	mSceneNode->setPosition(_position);
}
// ============================================================================
void Missile::setOrientation(Ogre::Quaternion & _orientation)
{
	mSceneNode->setOrientation(_orientation);
}
// ============================================================================
void Missile::processAccceleration(unsigned long timeSinceInMilis, float multiplier)
{
	Ogre::Vector3 globalFacingDirection = mSceneNode->getOrientation() * Ogre::Vector3::UNIT_Z;
	globalFacingDirection.normalise();

	Ogre::Vector3 globalAcceleration = multiplier * globalFacingDirection * 
								mMaxAcceleration * (float)(timeSinceInMilis)/1000;

	mAcceleration +=  globalAcceleration;

	if(mAcceleration.length() > mMaxAcceleration)
	{
		mAcceleration.normalise();
		mAcceleration *= mMaxAcceleration;
	}
}
// ============================================================================
void Missile::processVelocity(unsigned long timeSinceInMilis)
{
	mVelocity += mAcceleration;
	
	if(mVelocity.length() >= mMaxSpeed)
	{
		mVelocity.normalise();
		mVelocity *= mMaxSpeed;
	}
}
// ============================================================================
void Missile::processPosition(unsigned long timeSinceInMilis)
{
	mSceneNode->translate(mVelocity);
}
// ============================================================================
bool Missile::initializeFromPacket( const char * networkOrderedBuffer, Ogre::String sceneManager )
{
	int cnt = 0;
	// skip PacketID
	cnt += sizeof(char);

	// Owner ID
	char owner[MAXPLAYERNAMESIZE];
	memset(owner,0,MAXPLAYERNAMESIZE);
	memcpy(owner,&networkOrderedBuffer[cnt],MAXPLAYERNAMESIZE);
	mOwner = std::string(owner);
	cnt += MAXPLAYERNAMESIZE;

	// Entity ID + mesh and SceneNode init
	char entityId[MAXENTITYIDSIZE];
	memset(entityId,0,MAXENTITYIDSIZE);
	memcpy(entityId,&networkOrderedBuffer[cnt],MAXENTITYIDSIZE);
	mId = std::string(entityId);
	cnt += MAXENTITYIDSIZE;

	if(!mMesh)
	{
		mMesh = Ogre::Root::getSingleton().getSceneManager(
			sceneManager)->createEntity(mId, "Missile.mesh");
	}

	if(!mSceneNode)
	{
		mSceneNode = Ogre::Root::getSingleton().getSceneManager(
			sceneManager)->getRootSceneNode()->createChildSceneNode(mId+"Node");

		mSceneNode->attachObject(mMesh);
		// Missile model scale
		mSceneNode->setScale( Ogre::Vector3(0.3,0.3,0.3) ); // was: Ogre::Vector3(0.15,0.15,0.15)
	}

	// Position
	Ogre::Vector3 position;
	ParsingUtils::parseNetworkedVector3ToHost(position,&networkOrderedBuffer[cnt]);
	cnt += 3*sizeof(float);
	mSceneNode->setPosition(position);

	// Orientation
	Ogre::Quaternion orientation;
	ParsingUtils::parseNetworkedQuaternionToHost(orientation,&networkOrderedBuffer[cnt]);
	cnt += 4*sizeof(float);
	mSceneNode->setOrientation(orientation);

	// Velocity
	ParsingUtils::parseNetworkedVector3ToHost(mVelocity,&networkOrderedBuffer[cnt]);
	cnt += 3*sizeof(float);

	// Acceleration
	ParsingUtils::parseNetworkedVector3ToHost(mAcceleration,&networkOrderedBuffer[cnt]);
	cnt += 3*sizeof(float);

	mIsInitialized = true;
	
	// No problems at all 
	return true;
}
// ============================================================================
bool Missile::writeInitializationDataToPacket( char * networkOrderedBuffer )
{
	int cnt = 0;

	networkOrderedBuffer[cnt] = ID_ServerCreatesMissile;
	cnt += sizeof(char);

	// Owner ID
	memset(&networkOrderedBuffer[cnt],0,MAXPLAYERNAMESIZE);
	memcpy(&networkOrderedBuffer[cnt], mOwner.c_str(), mOwner.length());
	cnt += MAXPLAYERNAMESIZE;

	// Entity ID
	memset(&networkOrderedBuffer[cnt],0,MAXENTITYIDSIZE);
	memcpy(&networkOrderedBuffer[cnt], mId.c_str(), mId.length());
	cnt += MAXENTITYIDSIZE;

	// Position
	ParsingUtils::parseHostVector3ToNetwork(&networkOrderedBuffer[cnt],
		Ogre::Vector3(mSceneNode->getPosition()));
	cnt += 3*sizeof(float);

	// Orientation
	ParsingUtils::parseHostQuaternionToNetwork(&networkOrderedBuffer[cnt],
		Ogre::Quaternion(mSceneNode->getOrientation()));
	cnt += 4*sizeof(float);
	
	// Velocity
	ParsingUtils::parseHostVector3ToNetwork(&networkOrderedBuffer[cnt],mVelocity);
	cnt += 3*sizeof(float);

	// Acceleration
	ParsingUtils::parseHostVector3ToNetwork(&networkOrderedBuffer[cnt],mAcceleration);
	cnt += 3*sizeof(float);

	return true;
}
// ============================================================================
void Missile::readDataFromPacket(const char * networkOrderedBuffer )
{
	int cnt = 0;
	// skip PacketID
	cnt += sizeof(char);

	char entID[MAXENTITYIDSIZE];
	memset(entID,0,MAXENTITYIDSIZE);
	memcpy(entID,&networkOrderedBuffer[cnt],MAXENTITYIDSIZE);
	std::string sEntId = std::string(entID);
	cnt += MAXENTITYIDSIZE;

	// Wrong receiver?
	if(mId.compare(sEntId) != 0)
		return;

	// Position
	Ogre::Vector3 position;
	ParsingUtils::parseNetworkedVector3ToHost(position,&networkOrderedBuffer[cnt]);
	cnt += 3*sizeof(float);
	mSceneNode->setPosition(position);

	// Orientation
	Ogre::Quaternion orientation;
	ParsingUtils::parseNetworkedQuaternionToHost(orientation,&networkOrderedBuffer[cnt]);
	cnt += 4*sizeof(float);
	mSceneNode->setOrientation(orientation);

	// Velocity
	ParsingUtils::parseNetworkedVector3ToHost(mVelocity,&networkOrderedBuffer[cnt]);
	cnt += 3*sizeof(float);

	// Acceleration
	ParsingUtils::parseNetworkedVector3ToHost(mAcceleration,&networkOrderedBuffer[cnt]);
	cnt += 3*sizeof(float);
}
// ============================================================================
void Missile::saveDataToPacket( char * networkOrderedBuffer )
{
	int cnt = 0;

	networkOrderedBuffer[cnt] = ID_MissileData;
	cnt += sizeof(char);

	// Enemy ID
	memset(&networkOrderedBuffer[cnt],0,MAXENTITYIDSIZE);
	memcpy(&networkOrderedBuffer[cnt], mId.c_str(), mId.length());
	cnt += MAXENTITYIDSIZE;

	// Position
	ParsingUtils::parseHostVector3ToNetwork(&networkOrderedBuffer[cnt],
		Ogre::Vector3(mSceneNode->getPosition()));
	cnt += 3*sizeof(float);

	// Orientation
	ParsingUtils::parseHostQuaternionToNetwork(&networkOrderedBuffer[cnt],
		Ogre::Quaternion(mSceneNode->getOrientation()));
	cnt += 4*sizeof(float);
	
	// Velocity
	ParsingUtils::parseHostVector3ToNetwork(&networkOrderedBuffer[cnt],mVelocity);
	cnt += 3*sizeof(float);

	// Acceleration
	ParsingUtils::parseHostVector3ToNetwork(&networkOrderedBuffer[cnt],mAcceleration);
	cnt += 3*sizeof(float);
}