
#include <Ogre.h>
#include <Winsock2.h>
#include <ctype.h>
#include <stdarg.h>
#include <string.h>

#include "stdint.h"
#include "ParsingUtils.h"

// ============================================================================
uint32_t ParsingUtils::htonf(float f)
{
    uint32_t p;
    uint32_t sign;

    if (f < 0) { sign = 1; f = -f; }
    else { sign = 0; }
        
    p = ((((uint32_t)f)&0x7fff)<<16) | (sign<<31); // whole part and sign
    p |= (uint32_t)(((f - (int)f) * 65536.0f))&0xffff; // fraction

    return p;
}
// ============================================================================
float ParsingUtils::ntohf(uint32_t p)
{
    float f = ((p>>16)&0x7fff); // whole part
    f += (p&0xffff) / 65536.0f; // fraction

    if (((p>>31)&0x1) == 0x1) { f = -f; } // sign bit set

    return f;
}
// ============================================================================
#define pack754_32(f) (ParsingUtils::pack754((f), 32, 8))
#define pack754_64(f) (ParsingUtils::pack754((f), 64, 11))
#define unpack754_32(i) (ParsingUtils::unpack754((i), 32, 8))
#define unpack754_64(i) (ParsingUtils::unpack754((i), 64, 11))

long long ParsingUtils::pack754(long double f, unsigned bits, unsigned expbits)
{
    long double fnorm;
    int shift;
    long long sign, exp, significand;
    unsigned significandbits = bits - expbits - 1; // -1 for sign bit

    if (f == 0.0) return 0; // get this special case out of the way

    // check sign and begin normalization
    if (f < 0) { sign = 1; fnorm = -f; }
    else { sign = 0; fnorm = f; }

    // get the normalized form of f and track the exponent
    shift = 0;
    while(fnorm >= 2.0) { fnorm /= 2.0; shift++; }
    while(fnorm < 1.0) { fnorm *= 2.0; shift--; }
    fnorm = fnorm - 1.0;

    // calculate the binary form (non-float) of the significand data
    significand = fnorm * ((1LL<<significandbits) + 0.5f);

    // get the biased exponent
    exp = shift + ((1<<(expbits-1)) - 1); // shift + bias

    // return the final answer
    return (sign<<(bits-1)) | (exp<<(bits-expbits-1)) | significand;
}
// ============================================================================
long double ParsingUtils::unpack754(long long i, unsigned bits, unsigned expbits)
{
    long double result;
    long long shift;
    unsigned bias;
    unsigned significandbits = bits - expbits - 1; // -1 for sign bit

    if (i == 0) return 0.0;

    // pull the significand
    result = (i&((1LL<<significandbits)-1)); // mask
    result /= (1LL<<significandbits); // convert back to float
    result += 1.0f; // add the one back on

    // deal with the exponent
    bias = (1<<(expbits-1)) - 1;
    shift = ((i>>significandbits)&((1LL<<expbits)-1)) - bias;
    while(shift > 0) { result *= 2.0; shift--; }
    while(shift < 0) { result /= 2.0; shift++; }

    // sign it
    result *= (i>>(bits-1))&1? -1.0: 1.0;

    return result;
}
// ============================================================================
/*
** packi16() -- store a 16-bit int into a char buffer (like htons())
*/ 
void ParsingUtils::packi16(unsigned char *buf, unsigned int i)
{
    *buf++ = i>>8; *buf++ = i;
}
// ============================================================================
/*
** packi32() -- store a 32-bit int into a char buffer (like htonl())
*/ 
void ParsingUtils::packi32(unsigned char *buf, unsigned long i)
{
    *buf++ = i>>24; *buf++ = i>>16;
    *buf++ = i>>8;  *buf++ = i;
}
// ============================================================================
/*
** unpacki16() -- unpack a 16-bit int from a char buffer (like ntohs())
*/ 
unsigned int ParsingUtils::unpacki16(unsigned char *buf)
{
    return (buf[0]<<8) | buf[1];
}
// ============================================================================
/*
** unpacki32() -- unpack a 32-bit int from a char buffer (like ntohl())
*/ 
unsigned long ParsingUtils::unpacki32(unsigned char *buf)
{
    return (buf[0]<<24) | (buf[1]<<16) | (buf[2]<<8) | buf[3];
}
// ============================================================================
/*
** pack() -- store data dictated by the format string in the buffer
**
**  h - 16-bit              l - 32-bit
**  c - 8-bit char          f - float, 32-bit
**  s - string (16-bit length is automatically prepended)
*/ 
size_t ParsingUtils::pack(unsigned char *buf, char *format, ...)
{
    va_list ap;
    int h;
    int l;
    char c;
    float f;
    char *s;
    size_t size = 0, len;

    va_start(ap, format);

    for(; *format != '\0'; format++) {
        switch(*format) {
        case 'h': // 16-bit
            size += 2;
            h = va_arg(ap, int); // promoted
            packi16(buf, h);
            buf += 2;
            break;

        case 'l': // 32-bit
            size += 4;
            l = va_arg(ap, int);
            packi32(buf, l);
            buf += 4;
            break;

        case 'c': // 8-bit
            size += 1;
            c = va_arg(ap, int); // promoted
            *buf++ = (c>>0)&0xff;
            break;

        case 'f': // float
            size += 4;
            f = va_arg(ap, double); // promoted
            l = pack754_32(f); // convert to IEEE 754
            packi32(buf, l);
            buf += 4;
            break;

        case 's': // string
            s = va_arg(ap, char*);
            len = strlen(s);
            size += len + 2;
            packi16(buf, len);
            buf += 2;
            memcpy(buf, s, len);
            buf += len;
            break;
        }
    }

    va_end(ap);

    return size;
}
// ============================================================================
/*
** unpack() -- unpack data dictated by the format string into the buffer
*/
void ParsingUtils::unpack(unsigned char *buf, char *format, ...)
{
    va_list ap;
    short *h;
    int *l;
    int pf;
    char *c;
    float *f;
    char *s;
    size_t len, count, maxstrlen=0;

    va_start(ap, format);

    for(; *format != '\0'; format++) {
        switch(*format) {
        case 'h': // 16-bit
            h = va_arg(ap, short*);
            *h = unpacki16(buf);
            buf += 2;
            break;

        case 'l': // 32-bit
            l = va_arg(ap, int*);
            *l = unpacki32(buf);
            buf += 4;
            break;

        case 'c': // 8-bit
            c = va_arg(ap, char*);
            *c = *buf++;
            break;

        case 'f': // float
            f = va_arg(ap, float*);
            pf = unpacki32(buf);
            buf += 4;
            *f = unpack754_32(pf);
            break;

        case 's': // string
            s = va_arg(ap, char*);
            len = unpacki16(buf);
            buf += 2;
            if (maxstrlen > 0 && len > maxstrlen) count = maxstrlen - 1;
            else count = len;
            memcpy(s, buf, count);
            s[count] = '\0';
            buf += len;
            break;

        default:
            if (isdigit(*format)) { // track max str len
                maxstrlen = maxstrlen * 10 + (*format-'0');
            }
        }

        if (!isdigit(*format)) maxstrlen = 0;
    }

    va_end(ap);
}
// ============================================================================
void ParsingUtils::parseNetworkedVector3ToHost(Ogre::Vector3 & vector, const char * buffer)
{
	//int counter = 0;
	//float tmpFloat = 0.0f;

	//memcpy(&tmpFloat,&buffer[counter],sizeof(float));
	//counter += sizeof(float);
	//vector.x = (float)ntohl((float)tmpFloat);

	//memcpy(&tmpFloat,&buffer[counter],sizeof(float));
	//counter += sizeof(float);
	//vector.y = (float)ntohl((float)tmpFloat);

	//memcpy(&tmpFloat,&buffer[counter],sizeof(float));
	////counter += sizeof(float);
	//vector.z = (float)ntohl((float)tmpFloat);

	unpack((unsigned char *) buffer, "fff", &vector.x, &vector.y, &vector.z);
}
// ============================================================================
void ParsingUtils::parseNetworkedQuaternionToHost(Ogre::Quaternion & quaternion, const char * buffer)
{
	//int counter = 0;
	//float tmpFloat = 0.0f;

	//memcpy(&tmpFloat,&buffer[counter],sizeof(float));
	//counter += sizeof(float);
	//quaternion.w = (float)ntohl(tmpFloat);

	//memcpy(&tmpFloat,&buffer[counter],sizeof(float));
	//counter += sizeof(float);
	//quaternion.x = (float)ntohl(tmpFloat);

	//memcpy(&tmpFloat,&buffer[counter],sizeof(float));
	//counter += sizeof(float);
	//quaternion.y = (float)ntohl(tmpFloat);

	//memcpy(&tmpFloat,&buffer[counter],sizeof(float));
	////counter += sizeof(float);
	//quaternion.z = (float)ntohl(tmpFloat);

	unpack((unsigned char *) buffer, "ffff", &quaternion.w, &quaternion.x, &quaternion.y, &quaternion.z);
}
// ============================================================================
void ParsingUtils::parseNetworkedFloatToHost(float & destFloat, const char * buffer)
{
	//float tmpFloat = 0.0f;
	//memcpy(&tmpFloat,buffer,sizeof(float));
	//destFloat = (float)ntohl((float)tmpFloat);

	unpack((unsigned char *)buffer,"f",&destFloat);
}
// ============================================================================
void ParsingUtils::parseHostFloatToNetwork(char * destBuffer, const float & sourceFloat)
{
	//float tmpFloat = htonl(sourceFloat);
	//memcpy(destBuffer,&tmpFloat,sizeof(float));

	pack((unsigned char *)destBuffer, "f", sourceFloat);
}
// ============================================================================
void ParsingUtils::parseHostQuaternionToNetwork(char * destBuffer, const Ogre::Quaternion & quaternion)
{
	//int counter = 0;
	//float tmpFloat = 0.0f;

	//tmpFloat = htonl(quaternion.w);
	//memcpy(&destBuffer[counter],&tmpFloat,sizeof(float));
	//counter += sizeof(float);

	//tmpFloat = htonl(quaternion.x);
	//memcpy(&destBuffer[counter],&tmpFloat,sizeof(float));
	//counter += sizeof(float);

	//tmpFloat = htonl(quaternion.y);
	//memcpy(&destBuffer[counter],&tmpFloat,sizeof(float));
	//counter += sizeof(float);

	//tmpFloat = htonl(quaternion.z);
	//memcpy(&destBuffer[counter],&tmpFloat,sizeof(float));
	////counter += sizeof(float);

	pack((unsigned char *)destBuffer,"ffff",quaternion.w, quaternion.x, quaternion.y, quaternion.z);
}
// ============================================================================
void ParsingUtils::parseHostVector3ToNetwork(char * destBuffer, const Ogre::Vector3 & vector)
{
	//int counter = 0;
	//float tmpFloat = 0.0f;

	//tmpFloat = htonl(vector.x);
	//float tmpFloat2 = pack754_32(
	//memcpy(&destBuffer[counter],&tmpFloat,sizeof(float));
	//counter += sizeof(float);

	//tmpFloat = htonl(vector.y);
	//memcpy(&destBuffer[counter],&tmpFloat,sizeof(float));
	//counter += sizeof(float);

	//tmpFloat = htonl(vector.z);
	//memcpy(&destBuffer[counter],&tmpFloat,sizeof(float));
	////counter += sizeof(float);

	// My crude functions are no match for Beej's :D

	pack((unsigned char *)destBuffer, "fff", vector.x, vector.y, vector.z);
}
// ============================================================================
